/*******************************************************************************
*
*	TITLE		quick_adc.c 
*
*	VERSION:	0.1 (Beta)                           
*
*	DATE:		02-Feb-2008
*
*	AUTHOR:		R. Kevin Watson
*				kevinw@jpl.nasa.gov
*
*	COMMENTS:	This software will work with all 2004-2008 IFI 
*				robot controllers.
*
*				This software is not compatible with the software
*				found in adc.c/.h.
*
*				This file best viewed with tabs set to four. 
*
*				You are free to use this source code for any non-commercial
*				use. Please do not make copies of this source code, modified
*				or un-modified, publicly available on the internet or
*				elsewhere without permission. Thanks.
*
*				Copyright 2008 R. Kevin Watson. All rights are reserved.
*
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	02-Feb-2008  0.1  RKW - Original code.
*
*******************************************************************************/

#include "ifi_frc.h"
#include "quick_adc.h"

/*******************************************************************************
*
*	FUNCTION:		Quick_ADC()
*
*	PURPOSE:		Does a single quick analog to digital conversion.
*
*	CALLED FROM:
*
*	PARAMETERS:		Unsigned char containing a value between one and
*					sixteen indicating the desired ADC channel.
*
*	RETURNS:		Unsigned int containing a value between 0 and 1024.
*
*	COMMENTS:		This software will work with all 2004-2008 IFI
*					robot controllers.
*
*					This software is not compatible with the software
*					found in adc.c/.h.
*
*******************************************************************************/
unsigned int Quick_ADC(unsigned char channel)  
{
	unsigned char adcon0_temp;
	unsigned int adc;
	unsigned char i;

	// decrement the channel number because channel
	// numbers start at zero, not one
	channel--;

	// make sure the ADC interrupt is disabled
	PIE1bits.ADIE = 0;

	// enable all sixteen analog inputs
	ADCON1 = 0;

	// select Fosc/64 as the ADC Conversion Clock, right justify data
	// in the conversion result register and select 20 Tad acquisition
	// period
	ADCON2 = 0b10111110;

	// Select the ADC channel. This also starts the process whereby the ADC 
	// sample and hold capacitor is allowed to start charging, which must be
	// completed before the analog to digital conversion can be started.
	adcon0_temp = channel;
	adcon0_temp <<= 2;
	adcon0_temp |= 0b00000001;
	ADCON0 = adcon0_temp;

	// a short delay to allow the ADC holding capacitor to charge
	for(i=0; i<20; i++);

	// start a new analog to digital conversion
	ADCON0bits.GO = 1;

	// wait for the conversion to finish
	while(ADCON0bits.GO == 1);

	// get the conversion result
	adc = ADRESH;
	adc <<= 8;
	adc += ADRESL;

	// turn off the ADC hardware
	ADCON0 = 0;

	return(adc);
}

/*******************************************************************************
*
*	FUNCTION:		Quick_ADC_to_mV()
*
*	PURPOSE:		Converts the raw output of the Quick_ADC()
*					function to millivolts.
*
*	CALLED FROM:
*
*	PARAMETERS:		Quick_ADC() output value to convert
*
*	RETURNS:		Unsigned int containing the value in millivolts
*
*	COMMENTS:
*
*******************************************************************************/
unsigned int Quick_ADC_to_mV(unsigned int adc)
{
	return((unsigned int)(((long)adc*625L)/128L));
}
